"""
文本滑动窗口工具类

由于openai接口最大单次请求只支持8000个token，因此超过8000个token的文章需要分段发送翻译。然而直接粗暴的按文本长度截断会导致破坏语义，甚至破坏单词。
因此需要按行、按句，根据累计的token数来做一个滑动窗口。

思路是用生成器的方式，产出文章片段。
"""
import nltk

nltk.download("punkt")


def segments(doc: str, max_tokens: int, language: str = "english"):
    """将大文本按max_tokens大小分片

    Args:
        doc (str): 文本
        max_tokens (int, optional): 单片最大tokens数.

    Yields:
        str: 单片文本
    """

    # 先将文章按行切割（切割后换行符没了）
    lines = doc.splitlines()

    cur_segment = ""
    total_token = 0

    for line in lines:
        tokens = nltk.word_tokenize(line)
        token_count = len(tokens)
        if total_token + token_count > max_tokens:
            # 将要超过最大token
            yield cur_segment + "\n"  # 返回之前缓存的片
            cur_segment = line + "\n"  # 当前行设置为当前片
            total_token = token_count
        else:
            # 没超过最大token数就继续累加
            cur_segment += line + "\n"
            total_token += token_count

    yield cur_segment



import openai
import nltk
from tqdm import tqdm
from pydantic import BaseModel, RootModel
from ganf.segment import segments
from ganf.config import GLOBAL_OPENAI_CONF, OpenAIConfig


INIT_PROMPT = """
你现在是一个{extension}文档翻译器。
你必须满足以下几点：
- 不要破坏{extension}语法。
- 不要将原文的标点符号转换成中文标点符号（非常重要）。
- 要区分术语和说明文本，不用翻译术语。
- "[链接名]:(..路径)"碰到这种格式不要修改英文冒号！
- 只翻译描述性语言文本，{extension}代码不要做任何修改。
- 不要破坏python sphinx文档相关语法标记。
- 翻译的结果要能通过sphinx linter
- rst\restructuredtext 超链接标记前后要加空格隔开
{prompts}

我下面会把整个{extension}发送给你，你按照原来文档的格式翻译成{locale}语言返回给我。

"""


class Message(BaseModel):
    role: str
    content: str


MessageList = RootModel[list[Message]]


class Translator:
    def __init__(self, openai_conf: OpenAIConfig | str = GLOBAL_OPENAI_CONF) -> None:
        if isinstance(openai_conf, str):
            self.openai_conf = OpenAIConfig.load(openai_conf)
        else:
            self.openai_conf = openai_conf

        openai.api_key = self.openai_conf.api_key
        openai.api_base = self.openai_conf.api_base
        openai.api_type = self.openai_conf.api_type
        openai.api_version = self.openai_conf.api_version

    def chat_completion(self, messages: MessageList):
        if openai.api_type == "azure":
            response = openai.ChatCompletion.create(
                deployment_id=self.openai_conf.deployment_id,
                model=self.openai_conf.model,
                messages=messages.model_dump(),
            )
        else:
            response = openai.ChatCompletion.create(
                model=self.openai_conf.model, messages=messages.model_dump()
            )

        txt = response["choices"][0]["message"]["content"]  # type: ignore
        return txt

    def translate(
        self, content: str, from_locale: str, to_locale: str, prompts: list[str]
    ):
        translated_content: list[str] = []

        segs = segments(
            content, max_tokens=self.openai_conf.max_tokens, language=from_locale
        )

        for seg in tqdm(segs, desc="翻译中..."):
            if not seg.strip():
                translated_content.append(seg)
                continue

            result = self.chat_completion(
                MessageList(
                    [
                        Message(
                            role="system",
                            content=INIT_PROMPT.format(
                                locale=to_locale,
                                extension=".md",
                                prompts=prompts,
                            ),
                        ),
                        Message(role="user", content=seg),
                    ]
                )
            )
            translated_content.append(result)

        return "".join(translated_content)

    def cost_calculate(self, content: str):
        tokens = nltk.word_tokenize(content)
        cost = len(tokens) * (self.openai_conf.cost_per_k_token / 1000)
        return cost, len(tokens)



import glob
import os
import shutil
import sys
import click
from pydantic import BaseModel
from typing import Any, Callable, List
from contextvars import ContextVar
from click import Context
from tqdm import tqdm
from ganf.config import (
    GANF_CONF,
    GLOBAL_OPENAI_CONF,
    GANF_DIR,
    META_CONF,
    OPENAI_CONF,
    GanfConfig,
    MetaConfig,
    OpenAIConfig,
)
from gitignore_parser import parse_gitignore
from ganf.translator import Translator
from ganf.util import read_file, write_file

IGNORE_FILE = ".ganfignore"

IsIgnoreFunction = Callable[[str], bool]

openai_conf_var = ContextVar[OpenAIConfig]("openai_conf_var")


class CostInfo(BaseModel):
    cost: float = 0
    tokens: int = 0


class OrderCommands(click.Group):
    def list_commands(self, ctx: Context) -> List[str]:
        return list(self.commands)


def openai_configurate(ctx: Context, param: click.Parameter, value: str) -> Any:
    if not os.path.exists(value):
        click.echo(f"配置文件{value}不存在，转为使用全局配置。")
        value = GLOBAL_OPENAI_CONF
        # 如果全局都不存在就转而去执行openai配置设置向导了
        if not os.path.exists(value):
            answer = click.confirm("没有配置openai.toml，是否现在配置？", default=True, abort=True)
            if answer:
                local = click.confirm("是否使用本地配置？(否将创建全局配置)", default=True, abort=True)
                ctx.invoke(openai, local=local)

            sys.exit()

    openai_conf = OpenAIConfig.load(value)
    return openai_conf


openai_config_option = click.option(
    "--openai",
    type=click.Path(exists=False),
    default=OPENAI_CONF,
    callback=openai_configurate,
    show_default=True,
    help="openai配置文件",
)


def ganf_configurate(ctx: Context, param: click.Parameter, value: str) -> Any:
    if not os.path.exists(value):
        click.echo("当前目录没有配置文件，请使用 `ganf init` 命令创建配置文件。")
        sys.exit()

    ctx.default_map = ctx.default_map or {}
    ganf_config = GanfConfig.load(value)
    return ganf_config


ganf_config_option = click.option(
    "-c",
    "--config",
    type=click.Path(exists=False),
    default=GANF_CONF,
    callback=ganf_configurate,
    show_default=True,
    help="配置文件",
)


def ignore_configure(ctx: Context, param: click.Parameter, value: str):
    if os.path.exists(value):
        ignore = parse_gitignore(value)

        def _inner_ignore(path: str):
            dirname = os.path.dirname(path)
            return ignore(dirname) or ignore(path)

        return _inner_ignore
    else:
        click.echo("没有.ignore文件，将对所有文档进行翻译。")
        return lambda path: False


ganfignore_option = click.option(
    "-i",
    "--ignore",
    type=click.Path(exists=False),
    default=IGNORE_FILE,
    callback=ignore_configure,
    show_default=True,
    help="忽略列表",
)


def get_meta_config(path: str):
    if os.path.exists(path):
        return MetaConfig.load(path)
    else:
        return MetaConfig(file_name=path)


@click.group(cls=OrderCommands)
def main():
    """
    Ganf 文档翻译工具
    """


def openai_conf_wizard(save_to: str):
    config = OpenAIConfig()
    data = config.model_dump()
    for k, v in data.items():
        v = click.prompt(k, default=v, show_default=True)
        data[k] = v

    config = OpenAIConfig.model_validate(data)
    config.save(save_to)
    click.echo(save_to)


@main.command()
@click.option("--local", "-l", default=False, flag_value=True, help="全局还是本地")
def openai(local: bool):
    """
    创建 openai配置文件
    """

    if not local:
        click.echo("正在配置全局配置文件...")
        if not os.path.exists(GLOBAL_OPENAI_CONF):
            os.makedirs(GANF_DIR, exist_ok=True)
        else:
            click.echo("全局配置文件已存在，请删除后再试。")
            click.echo(GLOBAL_OPENAI_CONF)
            sys.exit()

        openai_conf_wizard(GLOBAL_OPENAI_CONF)
    else:
        click.echo("正在配置项目配置文件...")
        if not os.path.exists(OPENAI_CONF):
            openai_conf_wizard(OPENAI_CONF)
        else:
            click.echo("项目配置文件已存在，请删除后再试。")
            sys.exit()


@main.command()
@openai_config_option
@ganf_config_option
@ganfignore_option
def cost(openai: OpenAIConfig, config: GanfConfig, ignore: IsIgnoreFunction):
    """
    计算ganf项目的翻译成本
    """
    # 这一步检查不是必须的，只是做个防御和方便类型检查
    if config.file_name is None or not os.path.exists(config.file_name):
        click.echo("ganf.toml不存在")
        sys.exit()

    # ganf.toml可能不是在当前目录里，因此要将工作目录切换到ganf.toml所在目录
    # ganf.toml所在目录绝对路径
    dirname = os.path.dirname(os.path.abspath(config.file_name))
    # 将当前工作目录切换到ganf.toml所在目录
    os.chdir(dirname)

    # 原文档路径
    source_dir = config.source_dir
    # 通配符递归原文档
    files = glob.glob(os.path.join(source_dir, "**", "*.*"), recursive=True)

    bar_locale = tqdm(config.to_locales, desc="所有语言", leave=True)

    total_cost = CostInfo()

    translator = Translator(openai)
    locale_cost_map: dict[str, CostInfo] = {}

    for locale in bar_locale:
        locale_cost_map[locale] = locale_cost = CostInfo()

        dist_dir = os.path.join(config.dist_dir, locale)
        meta_conf_path = os.path.join(dist_dir, META_CONF)
        meta_config = get_meta_config(meta_conf_path)

        bar_locale.set_postfix_str(locale)

        # 过滤掉葫芦和已翻译文件
        def _filter(path: str):
            if ignore(path):  # 忽略
                return False

            if meta_config.is_modified(path):  # 被修改
                return True
            else:
                return False

        bar_calc = tqdm(filter(_filter, files), desc="计算...", leave=True)

        for file_path in bar_calc:
            content = read_file(file_path)
            c, t = translator.cost_calculate(content)
            locale_cost.cost += c
            locale_cost.tokens += t

        total_cost.cost += locale_cost.cost
        total_cost.tokens += locale_cost.tokens

    click.echo(f"总翻译语言: {len(config.to_locales)}")
    click.echo(f"总费用: ${total_cost.cost:.3f}")
    click.echo(f"总token数: {total_cost.tokens}")

    click.echo("-" * 10)

    for locale, cost in locale_cost_map.items():
        click.echo(f"{locale} cost:${cost.cost:.3f} tokens:{cost.tokens}")


@main.command()
@openai_config_option
@ganf_config_option
@ganfignore_option
def build(openai: OpenAIConfig, config: GanfConfig, ignore: IsIgnoreFunction):
    """
    构建ganf翻译
    """

    if config.file_name is None or not os.path.exists(config.file_name):
        click.echo("ganf.toml不存在")
        sys.exit()

    # ganf.toml可能不是在当前目录里，因此要将工作目录切换到ganf.toml所在目录
    # ganf.toml所在目录绝对路径
    dirname = os.path.dirname(os.path.abspath(config.file_name))
    # 将当前工作目录切换到ganf.toml所在目录
    os.chdir(dirname)

    # 原文档绝对路径
    source_dir = config.source_dir

    # 通配符递归原文档
    files = glob.glob(os.path.join(source_dir, "**", "*.*"), recursive=True)

    translator = Translator(openai)

    def _build():
        bar_locale = tqdm(config.to_locales, desc="所有语言", leave=True)

        for locale in bar_locale:
            dist_dir = os.path.join(config.dist_dir, locale)

            meta_conf_path = os.path.join(dist_dir, META_CONF)

            meta_config = get_meta_config(meta_conf_path)
            bar_locale.set_postfix_str(locale)

            def _filter(path: str):
                if ignore(path):  # 忽略
                    return False

                if meta_config.is_modified(path):  # 被修改
                    return True
                else:
                    return False

            bar_tran = tqdm(files, desc=locale, leave=True)

            for file_path in bar_tran:
                save_to = file_path.replace(source_dir, dist_dir)

                if not os.path.exists(save_to):
                    # 目标目录没有这个文件，那么就拷贝过去
                    dist_dir = os.path.dirname(save_to)
                    if not os.path.exists(dist_dir):
                        os.makedirs(dist_dir, exist_ok=True)
                    shutil.copy(file_path, save_to)

                if _filter(file_path) is False:
                    continue
                else:
                    content = read_file(file_path)
                    translated_content = translator.translate(
                        content, config.from_locale, locale, config.prompts
                    )
                    # 替换输出目录，保持目录结构
                    write_file(translated_content, save_to)

                    # 更新md5，并保存
                    meta_config.update(file_path)
                    meta_config.save()

    _build()



import textwrap


class OutOfMaxTokensError(Exception):
    def __init__(
        self, line_no: int, content: str, max_tokens: int, *args: object
    ) -> None:
        super().__init__(
            f"line:{line_no} 句子'{textwrap.shorten(content,width=20,placeholder='...')}'的token数量超过了{max_tokens}",
            *args,
        )



import json
from typing import Any, Literal
import openai
import toml
import os
from collections import defaultdict
from pydantic import BaseModel, Field
from contextvars import ContextVar
from .util import file_md5

USER_DIR = os.path.expanduser("~")
GANF_DIR = os.path.join(USER_DIR, ".ganf")

OPENAI_CONF = "openai.toml"
GLOBAL_OPENAI_CONF = os.path.join(GANF_DIR, OPENAI_CONF)

GANF_CONF = "ganf.toml"

META_CONF = "meta.json"

GANFIGNORE_FILE = ".ganfignore"


def default_factory(env_var: str, default=None) -> Any:
    return lambda: os.environ.get(env_var, default)


class JsonConfig(BaseModel):
    file_name: str | None = Field(default=None, description="配置文件名", exclude=True)

    @classmethod
    def load(cls, file_name: str):
        """从配置文件加载。加载后将 `file_name` 赋值给 `self.file_name`。

        Args:
            file_name (str): _description_

        Returns:
            _type_: _description_
        """
        with open(file_name, encoding="utf-8") as f:
            data = json.load(f)

        config = cls(file_name=file_name, **data)
        return config

    def save(self, file_name: str | None = None):
        """保存文件，默认保存回 `load` 时的文件。

        Args:
            file_name (str | None, optional): 文件名. Defaults to None.

        Raises:
            ValueError: 文件名错误时
        """
        if file_name:
            with open(file_name, "w", encoding="utf-8") as f:
                json.dump(self.model_dump(), f, ensure_ascii=False, indent=4)
            self.file_name = file_name
        elif self.file_name:
            with open(self.file_name, "w", encoding="utf-8") as f:
                json.dump(self.model_dump(), f, ensure_ascii=False, indent=4)
        else:
            raise ValueError("file_name is None")


class TomlConfig(BaseModel):
    file_name: str | None = Field(default=None, description="配置文件名", exclude=True)

    @classmethod
    def load(cls, file_name: str):
        """从配置文件加载。加载后将 `file_name` 赋值给 `self.file_name`。

        Args:
            file_name (str): _description_

        Returns:
            _type_: _description_
        """
        with open(file_name, encoding="utf-8") as f:
            data = toml.load(f)

        config = cls(file_name=file_name, **data)
        return config

    def save(self, file_name: str | None = None):
        """保存文件，默认保存回 `load` 时的文件。

        Args:
            file_name (str | None, optional): 文件名. Defaults to None.

        Raises:
            ValueError: 文件名错误时
        """
        if file_name:
            with open(file_name, "w", encoding="utf-8") as f:
                toml.dump(self.model_dump(), f)
            self.file_name = file_name
        elif self.file_name:
            with open(self.file_name, "w", encoding="utf-8") as f:
                toml.dump(self.model_dump(), f)
        else:
            raise ValueError("file_name is None")


class OpenAIConfig(TomlConfig):
    """
    OpenAI配置文件
    """

    api_key: str = Field(
        default_factory=default_factory("OPENAI_API_KEY", openai.api_key),
        description="OpenAI API Key",
    )
    api_base: str = Field(
        default_factory=default_factory("OPENAI_API_BASE", openai.api_base),
        description="OpenAI API Base",
    )
    api_type: str | Literal["open_ai", "azure"] = Field(
        default_factory=default_factory("OPENAI_API_TYPE", openai.api_type),
        description="OpenAI API Type",
    )
    api_version: str = Field(
        default_factory=default_factory("OPENAI_API_VERSION", openai.api_version),
        description="OpenAI API Version",
    )
    model: str = Field(
        default_factory=default_factory("OPENAI_MODEL", "gpt-3.5-turbo"),
        description="OpenAI Model",
    )
    deployment_id: str = Field(
        default_factory=default_factory("OPENAI_DEPLOYMENT_ID"),
        description="OpenAI Deployment ID (Azure Need)",
    )
    RPM: int = Field(default=10, description="OpenAI RPM")
    max_tokens: int = Field(default=1000, description="OpenAI Max Tokens")
    cost_per_k_token: float = Field(
        default=0.0015, description="OpenAI Cost Per K Token"
    )


class GanfConfig(TomlConfig):
    """
    Ganf项目配置文件
    """

    source_dir: str = Field(default="docs", description="Source Directory")
    dist_dir: str = Field(default="dist", description="Dist Directory")
    from_locale: str = Field(default="en", description="From Locale")
    to_locales: list[str] = Field(default=["zh"], description="Locales")
    prompts: list[str] = Field(default_factory=list, description="Prompts")


class MetaConfig(JsonConfig):
    """
    记录了翻译进度，文件哈希值的文件
    """

    root: dict[str, str] = Field(
        default_factory=lambda: defaultdict(lambda: None), description="Root"
    )

    def update(self, path: str):
        self.root[path] = file_md5(path)

    def is_modified(self, path: str):
        """检测文件是否被修改过

        Args:
            path (str): 文件路径

        Returns:
            bool: True修改过，False未修改
        """
        if self.root[path]:
            return file_md5(path) != self.root[path]
        return True


openai_conf_var = ContextVar[OpenAIConfig]("openai_conf_var")



from hashlib import md5


def file_md5(file_path: str):
    """
    计算文件的md5值
    这个实现目前不支持超大文件（几个G那种）
    :param file_path: 文件路径
    :return: md5值
    """
    with open(file_path, "rb") as f:
        md5_obj = md5()
        md5_obj.update(f.read())
        md5_value = md5_obj.hexdigest()
        return md5_value


def read_file(path: str):
    with open(path, "r", encoding="utf-8") as f:
        return f.read()


def write_file(content: str, path: str):
    with open(path, "w", encoding="utf-8") as f:
        f.write(content)
